from fsspec_utils.utils.datetime import *
# import datetime as dt
# import re
# from functools import lru_cache

# import pendulum as pdl
# import polars as pl
# import polars.selectors as cs
# import pyarrow as pa


# def get_timestamp_column(df: pl.DataFrame | pl.LazyFrame | pa.Table) -> str | list[str]:
#     if isinstance(df, pa.Table):
#         df = pl.from_arrow(df).lazy()

#     # if isinstance(df, pl.LazyFrame):
#     #    return df.collect_schema().names()

#     return df.select(cs.datetime() | cs.date()).collect_schema().names()


# def get_timedelta_str(timedelta_string: str, to: str = "polars") -> str:
#     polars_timedelta_units = [
#         "ns",
#         "us",
#         "ms",
#         "s",
#         "m",
#         "h",
#         "d",
#         "w",
#         "mo",
#         "y",
#     ]
#     duckdb_timedelta_units = [
#         "nanosecond",
#         "microsecond",
#         "millisecond",
#         "second",
#         "minute",
#         "hour",
#         "day",
#         "week",
#         "month",
#         "year",
#     ]

#     unit = re.sub("[0-9]", "", timedelta_string).strip()
#     val = timedelta_string.replace(unit, "").strip()
#     if to == "polars":
#         return (
#             timedelta_string
#             if unit in polars_timedelta_units
#             else val
#             + dict(zip(duckdb_timedelta_units, polars_timedelta_units))[
#                 re.sub("s$", "", unit)
#             ]
#         )

#     if unit in polars_timedelta_units:
#         return (
#             f"{val} " + dict(zip(polars_timedelta_units, duckdb_timedelta_units))[unit]
#         )

#     return f"{val} " + re.sub("s$", "", unit)


# @lru_cache(maxsize=128)
# def timestamp_from_string(
#     timestamp: str,
#     tz: str | None = None,
#     exact: bool = True,
#     strict: bool = False,
#     naive: bool = False,
# ) -> pdl.DateTime | pdl.Date | pdl.Time | dt.datetime | dt.date | dt.time:
#     """
#     Converts a string like "2023-01-01 10:00:00" into a datetime.datetime object.

#     Args:
#         string (str): The string representation of the timestamp, e.g. "2023-01-01 10:00:00".
#         tz (str, optional): The timezone to use for the timestamp. Defaults to None.
#         exact (bool, optional): Whether to use exact parsing. Defaults to True.
#         strict (bool, optional): Whether to use strict parsing. Defaults to False.
#         naive (bool, optional): Whether to return a naive datetime without a timezone. Defaults to False.

#     Returns:
#         datetime.datetime: The datetime object.
#     """
#     # Extract the timezone from the string if not provided
#     # tz = extract_timezone(timestamp) if tz is None else tz
#     # timestamp = timestamp.replace(tz, "").strip() if tz else timestamp

#     pdl_timestamp = pdl.parse(timestamp, exact=exact, strict=strict)

#     if isinstance(pdl_timestamp, pdl.DateTime):
#         if tz is not None:
#             pdl_timestamp = pdl_timestamp.naive().set(tz=tz)
#         if naive or tz is None:
#             pdl_timestamp = pdl_timestamp.naive()

#     return pdl_timestamp


# def timedelta_from_string(
#     timedelta_string: str, as_timedelta
# ) -> pdl.Duration | dt.timedelta:
#     """
#     Converts a string like "2d10s" into a datetime.timedelta object.

#     Args:
#         string (str): The string representation of the timedelta, e.g. "2d10s".

#     Returns:
#         datetime.timedelta: The timedelta object.
#     """
#     # Extract the numeric value and the unit from the string
#     matches = re.findall(r"(\d+)([a-zA-Z]+)", timedelta_string)
#     if not matches:
#         raise ValueError("Invalid timedelta string")

#     # Initialize the timedelta object
#     delta = pdl.duration()

#     # Iterate over each match and accumulate the timedelta values
#     for value, unit in matches:
#         # Map the unit to the corresponding timedelta attribute
#         unit_mapping = {
#             "us": "microseconds",
#             "ms": "milliseconds",
#             "s": "seconds",
#             "m": "minutes",
#             "h": "hours",
#             "d": "days",
#             "w": "weeks",
#             "mo": "months",
#             "y": "years",
#         }
#         if unit not in unit_mapping:
#             raise ValueError("Invalid timedelta unit")

#         # Update the timedelta object
#         kwargs = {unit_mapping[unit]: int(value)}
#         delta += pdl.duration(**kwargs)

#     return delta.as_timedelta if as_timedelta else delta


# # def extract_timezone(timestamp_string):
# #     """
# #     Extracts the timezone from a timestamp string.

# #     Args:
# #         timestamp_string (str): The input timestamp string.

# #     Returns:
# #         str: The extracted timezone.
# #     """
# #     pattern = r"\b([a-zA-Z]+/{0,1}[a-zA-Z_ ]*)\b"  # Matches the timezone portion
# #     match = re.search(pattern, timestamp_string)
# #     if match:
# #         timezone = match.group(0)
# #         return timezone
# #     else:
# #         return None



from fsspec_utils.utils.misc import run_parallel
import posixpath
import re

import pyarrow as pa
import pyarrow.parquet as pq
import tqdm
from fsspec import filesystem as fsspec_filesystem, AbstractFileSystem
from joblib import Parallel, delayed

# # from ..schema import convert_large_types_to_normal
# from .datetime import timestamp_from_string
from .polars import pl

# # Compile regex patterns once for efficiency
# SPLIT_PATTERN = re.compile(
#     r"<=|<|>=|>|=|!=|\s+[n,N][o,O][t,T]\s+[i,I][n,N]\s+|\s+[i,I][n,N]\s+|\s+[i,I][s,S]\s+[n,N][o,O][t,T]\s+"
#     r"[n,N][u,U][l,L]{2}\s+|\s+[i,I][s,S]\s+[n,N][u,U][l,L]{2}\s+"
# )
# LOGICAL_OPERATORS_PATTERN = re.compile(
#     r"\s+[a,A][n,N][d,D] [n,N][o,O][t,T]\s+|\s+[a,A][n,N][d,D]\s+|\s+[o,O][r,R] [n,N][o,O][t,T]\s+|\s+[o,O][r,R]\s+"
# )


# # def sql2pyarrow_filter(string: str, schema: pa.Schema):
# #     """
# #     Generates a filter expression for PyArrow based on a given string and schema.

# #     Parameters:
# #         string (str): The string containing the filter expression.
# #         schema (pa.Schema): The PyArrow schema used to validate the filter expression.

# #     Returns:
# #         PyArrow.Expression: The generated filter expression.

# #     """

# #     def _parse_part(part):
# #         split_pattern = (
# #             r"<=|"
# #             r"<|"
# #             r">=|"
# #             r">|"
# #             r"=|"
# #             r"!=|"
# #             r"\s+[n,N][o,O][t,T]\s+[i,I][n,N]\s+|"
# #             r"\s+[i,I][n,N]\s+|"
# #             r"\s+[i,I][s,S]\s+[n,N][o,O][t,T]\s+[n,N][u,U][l,L]{2}\s+|"
# #             r"\s+[i,I][s,S]\s+[n,N][u,U][l,L]{2}\s+"
# #         )
# #         sign = re.findall(split_pattern, part)[0]
# #         # print(sign)
# #         # if "<" in sign or ">" in sign or "=" in sign or "!" in sign:
# #         field, val = [p.strip() for p in re.split(f"\\s*{sign}\\s*", part)]
# #         # else:
# #         #    field, val = [p.strip() for p in re.split(f"\s+{sign}\s+", part)]

# #         # print(field, val)
# #         type_ = schema.field(field).type
# #         if "(" in val:
# #             val = eval(val)

# #         if pa.types.is_time(type_):
# #             val = timestamp_from_string(val, exact=True)
# #             if isinstance(val, pdl.DateTime):
# #                 val = val.time()

# #         elif pa.types.is_date(type_):
# #             val = timestamp_from_string(val, exact=True)
# #             if isinstance(val, pdl.DateTime):
# #                 val = val.date()

# #         elif pa.types.is_timestamp(type_):
# #             val = timestamp_from_string(val, exact=True)

# #         elif pa.types.is_integer_value(type_) or pa.types.is_integer(type_):
# #             if isinstance(val, str):
# #                 val = int(float(val.strip("'").replace(",", ".")))
# #             elif isinstance(val, tuple):
# #                 val = tuple(
# #                     [
# #                         (
# #                             int(float(val_.strip(",").replace(",", ".")))
# #                             if isinstance(val_, str)
# #                             else val_
# #                         )
# #                         for val_ in val
# #                     ]
# #                 )
# #             elif isinstance(val, list):
# #                 val = [
# #                     (
# #                         int(float(val_.strip("'").replace(",", ".")))
# #                         if isinstance(val_, str)
# #                         else val_
# #                     )
# #                     for val_ in val
# #                 ]

# #         elif pa.types.is_float_value(type_) or pa.types.is_floating(type_):
# #             if isinstance(val, str):
# #                 val = float(val.strip("'").replace(",", "."))
# #             elif isinstance(val, tuple):
# #                 val = tuple(
# #                     [
# #                         (
# #                             float(val_.strip(",").replace(",", "."))
# #                             if isinstance(val_, str)
# #                             else val_
# #                         )
# #                         for val_ in val
# #                     ]
# #                 )
# #             elif isinstance(val, list):
# #                 val = [
# #                     (
# #                         float(val_.strip("'").replace(",", "."))
# #                         if isinstance(val_, str)
# #                         else val_
# #                     )
# #                     for val_ in val
# #                 ]

# #         elif pa.types.is_boolean(type_):
# #             val = bool(val.strip("'").replace(",", "."))

# #         if isinstance(val, str):
# #             val = val.strip("'")

# #         if sign == ">=":
# #             return pc.field(field) >= val
# #         elif sign == ">":
# #             return pc.field(field) > val
# #         elif sign == "<=":
# #             return pc.field(field) <= val
# #         elif sign == "<":
# #             return pc.field(field) < val
# #         elif sign == "=":
# #             return pc.field(field) == val
# #         elif sign == "!=":
# #             return pc.field(field) != val
# #         elif sign.lower() == "in":
# #             return pc.field(field).isin(val)
# #         elif sign.lower() == "not in":
# #             return ~pc.field(field).isin(val)
# #         elif sign.lower() == "is null":
# #             return pc.field(field).is_null(nan_is_null=True)
# #         elif sign.lower() == "is not null":
# #             return ~pc.field(field).is_null(nan_is_null=True)

# #     parts = re.split(
# #         (
# #             r"\s+[a,A][n,N][d,D] [n,N][o,O][t,T]\s+|"
# #             r"\s+[a,A][n,N][d,D]\s+|"
# #             r"\s+[o,O][r,R] [n,N][o,O][t,T]\s+|"
# #             r"\s+[o,O][r,R]\s+"
# #         ),
# #         string,
# #     )
# #     operators = [
# #         op.strip()
# #         for op in re.findall(
# #             (
# #                 r"\s+[a,A][n,N][d,D]\s+[n,N][o,O][t,T]\s+|"
# #                 r"\s+[a,A][n,N][d,D]\s+|"
# #                 r"[o,O][r,R]\s+[n,N][o,O][t,T]\s+|"
# #                 r"\s+[o,O][r,R]\s+"
# #             ),
# #             string,
# #         )
# #     ]

# #     # print(parts, operators)

# #     if len(parts) == 1:
# #         return _parse_part(parts[0])

# #     expr = _parse_part(parts[0])

# #     for part, operator in zip(parts[1:], operators):
# #         if operator.lower() == "and":
# #             expr = expr & _parse_part(part)
# #         elif operator.lower() == "and not":
# #             expr = expr & ~_parse_part(part)
# #         elif operator.lower() == "or":
# #             expr = expr | _parse_part(part)
# #         elif operator.lower() == "or not":
# #             expr = expr | ~_parse_part(part)

# #     return expr


# def sql2pyarrow_filter(string: str, schema: pa.Schema) -> pc.Expression:
#     """
#     Generates a filter expression for PyArrow based on a given string and schema.

#     Parameters:
#         string (str): The string containing the filter expression.
#         schema (pa.Schema): The PyArrow schema used to validate the filter expression.

#     Returns:
#         pc.Expression: The generated filter expression.

#     Raises:
#         ValueError: If the input string is invalid or contains unsupported operations.
#     """

#     def parse_value(val: str, type_: pa.DataType) -> Any:
#         """Parse and convert value based on the field type."""
#         if isinstance(val, (tuple, list)):
#             return type(val)(parse_value(v, type_) for v in val)

#         if pa.types.is_timestamp(type_):
#             return timestamp_from_string(val, exact=False, tz=type_.tz)
#         elif pa.types.is_date(type_):
#             return timestamp_from_string(val, exact=True).date()
#         elif pa.types.is_time(type_):
#             return timestamp_from_string(val, exact=True).time()

#         elif pa.types.is_integer(type_):
#             return int(float(val.strip("'").replace(",", ".")))
#         elif pa.types.is_floating(type_):
#             return float(val.strip("'").replace(",", "."))
#         elif pa.types.is_boolean(type_):
#             return val.lower().strip("'") in ("true", "1", "yes")
#         else:
#             return val.strip("'")

#     def _parse_part(part: str) -> pc.Expression:
#         match = SPLIT_PATTERN.search(part)
#         if not match:
#             raise ValueError(f"Invalid condition: {part}")

#         sign = match.group().lower().strip()
#         field, val = [p.strip() for p in SPLIT_PATTERN.split(part)]

#         if field not in schema.names:
#             raise ValueError(f"Unknown field: {field}")

#         type_ = schema.field(field).type
#         val = parse_value(val, type_)

#         operations = {
#             ">=": lambda f, v: pc.field(f) >= v,
#             ">": lambda f, v: pc.field(f) > v,
#             "<=": lambda f, v: pc.field(f) <= v,
#             "<": lambda f, v: pc.field(f) < v,
#             "=": lambda f, v: pc.field(f) == v,
#             "!=": lambda f, v: pc.field(f) != v,
#             "in": lambda f, v: pc.field(f).isin(v),
#             "not in": lambda f, v: ~pc.field(f).isin(v),
#             "is null": lambda f, v: pc.field(f).is_null(nan_is_null=True),
#             "is not null": lambda f, v: ~pc.field(f).is_null(nan_is_null=True),
#         }

#         if sign not in operations:
#             raise ValueError(f"Unsupported operation: {sign}")

#         return operations[sign](field, val)

#     parts = LOGICAL_OPERATORS_PATTERN.split(string)
#     operators = [op.lower().strip() for op in LOGICAL_OPERATORS_PATTERN.findall(string)]

#     if len(parts) == 1:
#         return _parse_part(parts[0])

#     expr = _parse_part(parts[0])
#     for part, operator in zip(parts[1:], operators):
#         if operator == "and":
#             expr = expr & _parse_part(part)
#         elif operator == "and not":
#             expr = expr & ~_parse_part(part)
#         elif operator == "or":
#             expr = expr | _parse_part(part)
#         elif operator == "or not":
#             expr = expr | ~_parse_part(part)
#         else:
#             raise ValueError(f"Unsupported logical operator: {operator}")

#     return expr


# def sql2polars_filter(string: str, schema: pl.Schema) -> pl.Expr:
#     """
#     Generates a filter expression for Polars based on a given string and schema.

#     Parameters:
#         string (str): The string containing the filter expression.
#         schema (pl.Schema): The Polars schema used to validate the filter expression.

#     Returns:
#         pl.Expr: The generated filter expression.

#     Raises:
#         ValueError: If the input string is invalid or contains unsupported operations.
#     """

#     def parse_value(val: str, dtype: pl.DataType) -> Any:
#         """Parse and convert value based on the field type."""
#         if isinstance(val, (tuple, list)):
#             return type(val)(parse_value(v, dtype) for v in val)

#         if dtype == pl.Datetime:
#             return timestamp_from_string(val, exact=False, tz=dtype.time_zone)
#         elif dtype == pl.Date:
#             return timestamp_from_string(val, exact=True).date()
#         elif dtype == pl.Time:
#             return timestamp_from_string(val, exact=True).time()
#         elif dtype in (pl.Int8, pl.Int16, pl.Int32, pl.Int64):
#             return int(float(val.strip("'").replace(",", ".")))
#         elif dtype in (pl.Float32, pl.Float64):
#             return float(val.strip("'").replace(",", "."))
#         elif dtype == pl.Boolean:
#             return val.lower().strip("'") in ("true", "1", "yes")
#         else:
#             return val.strip("'")

#     def _parse_part(part: str) -> pl.Expr:
#         match = SPLIT_PATTERN.search(part)
#         if not match:
#             raise ValueError(f"Invalid condition: {part}")

#         sign = match.group().lower().strip()
#         field, val = [p.strip() for p in SPLIT_PATTERN.split(part)]

#         if field not in schema.names():
#             raise ValueError(f"Unknown field: {field}")

#         dtype = schema[field]
#         val = parse_value(val, dtype)

#         operations = {
#             ">=": lambda f, v: pl.col(f) >= v,
#             ">": lambda f, v: pl.col(f) > v,
#             "<=": lambda f, v: pl.col(f) <= v,
#             "<": lambda f, v: pl.col(f) < v,
#             "=": lambda f, v: pl.col(f) == v,
#             "!=": lambda f, v: pl.col(f) != v,
#             "in": lambda f, v: pl.col(f).is_in(v),
#             "not in": lambda f, v: ~pl.col(f).is_in(v),
#             "is null": lambda f, v: pl.col(f).is_null(),
#             "is not null": lambda f, v: pl.col(f).is_not_null(),
#         }

#         if sign not in operations:
#             raise ValueError(f"Unsupported operation: {sign}")

#         return operations[sign](field, val)

#     parts = LOGICAL_OPERATORS_PATTERN.split(string)
#     operators = [op.lower().strip() for op in LOGICAL_OPERATORS_PATTERN.findall(string)]

#     if len(parts) == 1:
#         return _parse_part(parts[0])

#     expr = _parse_part(parts[0])
#     for part, operator in zip(parts[1:], operators):
#         if operator == "and":
#             expr = expr & _parse_part(part)
#         elif operator == "and not":
#             expr = expr & ~_parse_part(part)
#         elif operator == "or":
#             expr = expr | _parse_part(part)
#         elif operator == "or not":
#             expr = expr | ~_parse_part(part)
#         else:
#             raise ValueError(f"Unsupported logical operator: {operator}")

#     return expr


def read_table(
    path: str,
    schema: pa.Schema | None = None,
    filesystem: AbstractFileSystem | None = None,
    partitioning: str | list[str] | None = None,
) -> pa.Table:
    """Loads the data for the given file path into a pyarrow table.

    Args:
        path (str): File path.
        schema (pa.Schema | None, optional): Pyarrow schema. Defaults to None.
        filesystem (AbstractFileSystem | None, optional): Filesystem. Defaults to None.
        partitioning (str | list[str] | None, optional): Partitioning of the file. Could be 'hive' for
            hive style partitioning or a list of column names. Defaults to None.

    Returns:
        pa.Table: Pyarrow table.
    """
    # sourcery skip: avoid-builtin-shadow
    if filesystem is None:
        filesystem = fsspec_filesystem("file")

    filesystem.invalidate_cache()

    table = pq.read_table(pa.BufferReader(filesystem.read_bytes(path)), schema=schema)

    if partitioning is not None:
        partitions = get_partitions_from_path(path, partitioning=partitioning)

        for key, values in partitions:
            if key in table.column_names:
                table = table.drop(key)
            table = table.append_column(
                field_=key, column=pa.array([values] * len(table))
            )
    if schema is not None:
        return table.cast(schema)

    return table


def run_parallel(
    func: callable,
    func_params: list[any],
    *args,
    n_jobs: int = -1,
    backend: str = "threading",
    verbose: bool = True,
    **kwargs,
) -> list[any]:
    """Runs a function for a list of parameters in parallel.

    Args:
        func (Callable): function to run in Parallelallel.
        func_params (list[any]): parameters for the function
        n_jobs (int, optional): Number of joblib workers. Defaults to -1.
        backend (str, optional): joblib backend. Valid options are
        `loky`,`threading`, `mutliprocessing` or `sequential`.  Defaults to "threading".

    Returns:
        list[any]: Function output.
    """
    if verbose:
        return Parallel(n_jobs=n_jobs, backend=backend)(
            delayed(func)(fp, *args, **kwargs) for fp in tqdm.tqdm(func_params)
        )

    else:
        return Parallel(n_jobs=n_jobs, backend=backend)(
            delayed(func)(fp, *args, **kwargs) for fp in func_params
        )


def get_partitions_from_path(
    path: str, partitioning: str | list[str] | None = None
) -> list[tuple]:
    """Get the dataset partitions from the file path.

    Args:
        path (str): File path.
        partitioning (str | list[str] | None, optional): Partitioning type. Defaults to None.

    Returns:
        list[tuple]: Partitions.
    """
    if "." in path:
        path = posixpath.dirname(path)

    parts = path.split("/")

    if isinstance(partitioning, str):
        if partitioning == "hive":
            return [tuple(p.split("=")) for p in parts if "=" in p]

        else:
            return [
                (partitioning, parts[0]),
            ]
    else:
        return list(zip(partitioning, parts[-len(partitioning) :]))


def humanize_size(size: int, unit="MB") -> float:
    "Human-readable size"
    if unit.lower() == "b":
        return round(size, 1)
    elif unit.lower() == "kb":
        return round(size / 1024, 1)
    elif unit.lower() == "mb":
        return round(size / 1024**2, 1)
    elif unit.lower() == "gb":
        return round(size / 1024**3, 1)
    elif unit.lower() == "tb":
        return round(size / 1024**4, 1)
    elif unit.lower() == "pb":
        return round(size / 1024**5, 1)


def humanized_size_to_bytes(size: str) -> float:
    "Human-readable size t bytes"
    unit = re.sub("[0-9 ]", "", size).lower()
    size = float(re.sub("[a-zA-Z ]", "", size))
    if unit == "b":
        return int(size)
    elif unit == "kb":
        return int(size * 1024)
    elif unit == "mb":
        return int(size * 1024**2)
    elif unit == "gb":
        return int(size * 1024**3)
    elif unit == "tb":
        return int(size * 1024**4)
    elif unit == "pb":
        return int(size * 1024**5)


def getattr_rec(obj, attr_str):
    """
    Recursively retrieves an attribute from an object based on a dot-separated string.

    Args:
        obj: The object from which to retrieve the attribute.
        attr_str: A dot-separated string representing the attribute to retrieve.

    Returns:
        The value of the attribute.

    Raises:
        AttributeError: If the attribute does not exist.
    """
    attrs = attr_str.split(".")
    for attr in attrs:
        obj = getattr(obj, attr)
    return obj


def setattr_rec(obj, attr_str, value):
    """
    Recursively sets the value of an attribute in an object.

    Args:
        obj (object): The object to set the attribute in.
        attr_str (str): The attribute string in dot notation (e.g., "attr1.attr2.attr3").
        value: The value to set the attribute to.

    Returns:
        None
    """
    attrs = attr_str.split(".")
    for attr in attrs[:-1]:
        obj = getattr(obj, attr)
    setattr(obj, attrs[-1], value)


def delattr_rec(obj, attr_str):
    """
    Recursively deletes an attribute from an object.

    Args:
        obj: The object from which to delete the attribute.
        attr_str: A string representing the attribute to be deleted.
                  The attribute can be nested using dot notation (e.g., 'attr1.attr2.attr3').

    Raises:
        AttributeError: If the attribute does not exist.

    Example:
        obj = SomeObject()
        attr_str = 'attr1.attr2.attr3'
        delattr_rec(obj, attr_str)
    """
    attrs = attr_str.split(".")
    for attr in attrs[:-1]:
        obj = getattr(obj, attr)
    delattr(obj, attrs[-1])


def get_nested_keys(d, parent_key=""):
    """
    Recursively retrieves all the nested keys from a dictionary.

    Args:
        d (dict): The dictionary to retrieve the keys from.
        parent_key (str, optional): The parent key to prepend to the nested keys. Defaults to "".

    Returns:
        list: A list of all the nested keys in the dictionary.
    """
    keys = []
    for k, v in d.items():
        new_key = f"{parent_key}.{k}" if parent_key else k
        keys.append(new_key)
        if isinstance(v, dict):
            keys.extend(get_nested_keys(v, new_key))
    return keys


def unify_schemas_pl(schemas: list[pa.Schema]) -> pl.Schema:
    """
    Unifies a list of Pyarrow schemas into a single schema.

    Args:
        schemas (list[pl.Schema]): List of Polars schemas.

    Returns:
        pa.Schema: Unified schema.
    """
    schema = (
        pl.concat(
            [
                # pl.from_arrow(pa.Table.from_pylist([], schema=schema))
                pl.from_arrow(schema.empty_table())
                for schema in schemas
            ],
            how="diagonal_relaxed",
        )
        .to_arrow()
        .schema
    )

    return schema



from dataclasses import dataclass, field

import duckdb
import yaml
from fsspec import AbstractFileSystem

from .filesystem import FileSystem


@dataclass
class Filesystem:
    bucket: str | None = None
    fs: AbstractFileSystem | None = None
    profile: str | None = None
    key: str | None = None
    endpoint_url: str | None = None
    secret: str | None = None
    token: str | None = None
    protocol: str | None = None
    cached: bool = False
    cache_storage = "~/.tmp"
    check_files: bool = False
    cache_check: int = 120
    expire_time: int = 24 * 60 * 60
    same_names: bool = False
    kwargs: dict | None = None

    def connect(self):
        return FileSystem(**self.__dict__)


@dataclass
class Table:
    path: str
    type: str
    format: str
    filesystem: str
    partitioning: str | None = None
    # partitioning_flavor: str = "hive" # hive, directory, filename
    timestamp_column: str | None = None
    write_args: dict = field(default_factory=dict)

    def _load_parquet(self):
        pass


@dataclass
class Catalog:
    filesystems: Dict[str, Filesystem] = field(default_factory=dict)
    tables: Dict[str, Table] = field(default_factory=dict)
    ddb_con: Optional[duckdb.DuckDBPyConnection] = None
    _catalog_path: str = "catalog.yaml"

    def __post_init__(self):
        if self.ddb_con is None:
            self.ddb_con = duckdb.connect()

    def load_catalog(self):
        with open(self._catalog_path, "r") as f:
            data = yaml.safe_load(f)

        self.filesystems = {
            name: Filesystem(**fs) for name, fs in data["filesystems"].items()
        }
        self.tables = {
            name: Table(name=name, **table) for name, table in data["tables"].items()
        }

    def save_catalog(self):
        data = {
            "filesystems": {name: vars(fs) for name, fs in self.filesystems.items()},
            "tables": {
                name: {k: v for k, v in vars(table).items() if k != "name"}
                for name, table in self.tables.items()
            },
        }
        with open(self._catalog_path, "w") as f:
            yaml.dump(data, f)

    def create_filesystem(self, name: str, protocol: str, bucket: str, path: str):
        self.filesystems[name] = Filesystem(protocol=protocol, bucket=bucket, path=path)
        self.save_catalog()

    def create_table(
        self, name: str, path: str, format: str, filesystem: str, **kwargs
    ):
        self.tables[name] = Table(
            name=name, path=path, format=format, filesystem=filesystem, **kwargs
        )
        self.save_catalog()

    def delete_table(self, name: str):
        del self.tables[name]
        self.save_catalog()

    def write_table(self, table_name: str, data, **kwargs):
        table = self.tables[table_name]
        fs = self.filesystems[table.filesystem]
        # Implement writing logic here based on the format and filesystem
        # You may need to add more methods or use external libraries for actual writing

    def read_table(self, table_name: str, **kwargs):
        table = self.tables[table_name]
        fs = self.filesystems[table.filesystem]
        # Implement reading logic here based on the format and filesystem
        # You may need to add more methods or use external libraries for actual reading

    def sql(self, query: str):
        # Register tables if not already registered
        for table_name, table in self.tables.items():
            if table_name not in self.ddb_con.execute("SHOW TABLES").fetchall():
                self.ddb_con.register(table_name, self.read_table(table_name))

        return self.ddb_con.execute(query)


# Usage example:
if __name__ == "__main__":
    # %%
    catalog = Catalog()
    catalog.load_catalog()

    catalog.create_filesystem("s3", "s3", "my-bucket", "s3://my-bucket/data")
    catalog.create_table(
        "my_table", "s3://my-bucket/data/my_table.parquet", "parquet", "s3"
    )

    # Write data to the table (you need to implement the actual writing logic)
    data = ...  # Your data here
    catalog.write_table("my_table", data)

    # Read data from the table (you need to implement the actual reading logic)
    read_data = catalog.read_table("my_table")

    # Execute SQL query
    result = catalog.sql("SELECT * FROM my_table LIMIT 10")



"""
Security tests for PyDala2
"""
import pytest
import tempfile
import os
from pydala.helpers.security import (
    escape_sql_identifier,
    escape_sql_literal,
    validate_partition_name,
    validate_partition_value,
    sanitize_filter_expression,
    validate_path,
    safe_join,
    sanitize_filename,
)


class TestSqlSecurity:
    """Test SQL security utilities."""
    
    def test_escape_sql_identifier(self):
        """Test SQL identifier escaping."""
        # Valid identifiers
        assert escape_sql_identifier("column_name") == '"column_name"'
        assert escape_sql_identifier("_column") == '"_column"'
        assert escape_sql_identifier("column123") == '"column123"'
        
        # Invalid identifiers should raise
        with pytest.raises(ValueError):
            escape_sql_identifier("")
        with pytest.raises(ValueError):
            escape_sql_identifier("123column")  # Starts with number
        with pytest.raises(ValueError):
            escape_sql_identifier("column-name")  # Invalid char
    
    def test_escape_sql_literal(self):
        """Test SQL literal escaping."""
        assert escape_sql_literal(None) == "NULL"
        assert escape_sql_literal("test") == "'test'"
        assert escape_sql_literal("test's") == "'test''s'"  # Escaped quote
        assert escape_sql_literal(123) == "123"
        assert escape_sql_literal(True) == "TRUE"
        assert escape_sql_literal(False) == "FALSE"
    
    def test_validate_partition_name(self):
        """Test partition name validation."""
        # Valid names
        assert validate_partition_name("year") is True
        assert validate_partition_name("month") is True
        assert validate_partition_name("day_of_month") is True
        
        # Invalid names
        assert validate_partition_name("") is False
        assert validate_partition_name("year/month") is False
        assert validate_partition_name("year;DROP TABLE") is False
        assert validate_partition_name("a" * 256) is False
    
    def test_validate_partition_value(self):
        """Test partition value validation."""
        # Valid values
        assert validate_partition_value("2023") is True
        assert validate_partition_value(2023) is True
        assert validate_partition_value(True) is True
        assert validate_partition_value(None) is True
        
        # Invalid values
        assert validate_partition_value("../../../etc/passwd") is False
        assert validate_partition_value("value\0withnull") is False
        assert validate_partition_value("value\nwithnewline") is False
        assert validate_partition_value("a" * 1025) is False
    
    def test_sanitize_filter_expression(self):
        """Test filter expression sanitization."""
        # Valid expressions
        assert sanitize_filter_expression("column > 100") == "column > 100"
        assert sanitize_filter_expression("column = 'test'") == "column = 'test'"
        
        # Dangerous content should be removed
        assert "-- comment" not in sanitize_filter_expression("column > 100 -- comment")
        assert "/*" not in sanitize_filter_expression("column > 100 /* comment */")
        
        # Unbalanced quotes should raise
        with pytest.raises(ValueError):
            sanitize_filter_expression("column = 'test")


class TestPathSecurity:
    """Test path security utilities."""
    
    def test_validate_path(self):
        """Test path validation."""
        # Valid paths
        assert validate_path("valid/path") == "valid/path"
        assert validate_path("file.txt") == "file.txt"
        
        # Invalid paths
        with pytest.raises(ValueError):
            validate_path("")
        with pytest.raises(ValueError):
            validate_path("../etc/passwd")
        with pytest.raises(ValueError):
            validate_path("..\windows\system32")
        with pytest.raises(ValueError):
            validate_path("/absolute/path")
        
        # With base path
        base = "/safe/base"
        assert validate_path("subdir", base) == "subdir"
        with pytest.raises(ValueError):
            validate_path("../../../etc/passwd", base)
    
    def test_safe_join(self):
        """Test safe path joining."""
        base = "/safe/base"
        
        # Valid joins
        assert safe_join(base, "subdir") == "/safe/base/subdir"
        assert safe_join(base, "file.txt") == "/safe/base/file.txt"
        
        # Invalid joins should raise
        with pytest.raises(ValueError):
            safe_join(base, "../etc/passwd")
        with pytest.raises(ValueError):
            safe_join(base, "", "../malicious")
    
    def test_sanitize_filename(self):
        """Test filename sanitization."""
        # Valid filenames
        assert sanitize_filename("document.txt") == "document.txt"
        assert sanitize_filename("data_file.csv") == "data_file.csv"
        
        # Invalid characters removed
        assert sanitize_filename("doc<>ument.txt") == "document.txt"
        assert sanitize_filename("file/name.txt") == "filename.txt"
        
        # Edge cases
        assert sanitize_filename("") == "unnamed"
        assert sanitize_filename("   .   ") == "unnamed"
        assert len(sanitize_filename("a" * 300)) <= 255


class TestCredentialHandling:
    """Test credential handling security."""
    
    def test_get_credentials_redaction(self):
        """Test credential redaction."""
        from pydala.filesystem import get_credentials_from_fsspec
        from unittest.mock import Mock
        
        # Mock filesystem with credentials
        mock_fs = Mock()
        mock_fs.protocol = ["s3"]
        mock_creds = Mock()
        mock_creds.access_key = "AKIAIOSFODNN7EXAMPLE"
        mock_creds.secret_key = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
        mock_creds.token = "session_token_value"
        mock_fs.s3 = Mock()
        mock_fs.s3._get_credentials.return_value = mock_creds
        mock_fs.s3._endpoint = Mock()
        mock_fs.s3._endpoint.host = "s3.amazonaws.com"
        
        # Test redaction (default)
        creds = get_credentials_from_fsspec(mock_fs, redact_secrets=True)
        assert "REDACTED" in creds["access_key"]
        assert "REDACTED" in creds["secret_key"]
        assert creds["endpoint_override"] == "s3.amazonaws.com"
        
        # Test non-redacted
        creds = get_credentials_from_fsspec(mock_fs, redact_secrets=False)
        assert creds["access_key"] == "AKIAIOSFODNN7EXAMPLE"
        assert creds["secret_key"] == "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"


if __name__ == "__main__":
    pytest.main([__file__, "-v"])